#include <Python.h>

//Declarations
static PyObject* idea_optimized_encrypt(PyObject *self, PyObject *args);
void process_block(unsigned short *block, unsigned short* keys);
static PyObject* idea_optimized_decrypt(PyObject *self, PyObject *args);
unsigned short mod_mul(unsigned short a, unsigned short b);
PyMODINIT_FUNC initidea_optimized(void);

/* Python method definitions
 */
static PyMethodDef IdeaMethods[] = {
	{"encrypt",  idea_optimized_encrypt, METH_VARARGS, "encrypt using c optimized algorithm"},
	{"decrypt",  idea_optimized_decrypt, METH_VARARGS, "decrypt using c optimized algorithm"},
	{NULL, NULL, 0, NULL}		/* Sentinel */
};

/* Module initialization
 */
PyMODINIT_FUNC initidea_optimized(void)
{
	//Initialize the module and inform python about the functions available
	Py_InitModule("idea_optimized", IdeaMethods);
}

/* Addition under modulo 2^16-1
 */
#define mod_add(a,b) ((a+b)&0xffff)

/* Efficient implementation of modified multiplication
 * By Handbook of Applied Cryptography, Note 7.104
 */
unsigned short mod_mul(unsigned short a, unsigned short b)
{
	unsigned int c;
	int r;
	if(a==0)
		return (0x10001-b)&0xffff;
	else if(b==0)
		return (0x10001-a)&0xffff;
	else
		c = a*b;
		r = (c&0xffff) - (c>>16);
		if(r<0)
			r = 0x10001+r;
		return r&0xffff;
}

/* IDEA encryption interface from Python
 * Supports CBC and ECB using padded message and generated key
 */
static PyObject* idea_optimized_encrypt(PyObject *self, PyObject *args)
{
	char* input;
	char* operation_mode;
	int inputlen;
	unsigned short keys[52];

	//Parse the parameter tuple
	if(!PyArg_ParseTuple(args, "s#s(HHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHH)", &input, &inputlen, &operation_mode, &keys[0], &keys[1], &keys[2], &keys[3], &keys[4], &keys[5], &keys[6], &keys[7], &keys[8], &keys[9], &keys[10], &keys[11], &keys[12], &keys[13], &keys[14], &keys[15], &keys[16], &keys[17], &keys[18], &keys[19], &keys[20], &keys[21], &keys[22], &keys[23], &keys[24], &keys[25], &keys[26], &keys[27], &keys[28], &keys[29], &keys[30], &keys[31], &keys[32], &keys[33], &keys[34], &keys[35], &keys[36], &keys[37], &keys[38], &keys[39], &keys[40], &keys[41], &keys[42], &keys[43], &keys[44], &keys[45], &keys[46], &keys[47], &keys[48], &keys[49], &keys[50], &keys[51])){
		PyErr_SetString(PyExc_ValueError, "Input must be a string and key as 52-tuple of 16bit integers");
		return NULL;
	}
	/* Note: This is kind of a hack, but it doesn't seem like there's any other less hack way
	 * to parse a python list/tuple to a C integer array.
	 */

	//Check input padding, it must be perfectly padded
	if(inputlen%4){
		PyErr_SetString(PyExc_ValueError, "Input must be padded to a multiple of 64bit");
		return NULL;
	}

	//Read operations mode
	short mode;
	if(strcmp(operation_mode,"ECB") == 0){
		mode = 0;
	}else if(strcmp(operation_mode,"CBC") == 0){
		mode = 4;
	}else{
		PyErr_SetString(PyExc_ValueError, "An operation mode ECB or CBC must be provided");
		return NULL;
	}

	//Build a python string as return value
	PyObject* output = PyString_FromStringAndSize(NULL, inputlen+mode*sizeof(unsigned short));
	/* Note: We could be using malloc instead of creating an uinitialized python string
	 * However, with this approach we don't have to create a new copy. We may only do this
	 * because we're the only one with a reference the string, thus nobody else can modify it.
	 */
	if(!output)
		return PyErr_NoMemory();	//Raise exception if out of memory

	//Get pointer to python string buffer
	unsigned short* blocks = (unsigned short*)PyString_AsString(output);

	//Copy input to temporary variable
	memcpy(blocks+mode, input, inputlen);

	//Add IV (initialization vector) if needed, using when running CBC mode
	if(mode){
		blocks[0] = 3;
		blocks[1] = 14;
		blocks[2] = 15;
		blocks[3] = 92;
	}

	//Encrypt blocks
	unsigned int i;
	for(i=mode; i<(inputlen/sizeof(unsigned short)+mode); i+=4){
		if(mode){
			blocks[i] ^= blocks[i-4];
			blocks[i+1] ^= blocks[i-3];
			blocks[i+2] ^= blocks[i-2];
			blocks[i+3] ^= blocks[i-1];
		}
		process_block(blocks+i, keys);
	}

	//Return encrypted binary python string
	return output;
}

/* IDEA decryption interface from Python
 * Supports CBC and ECB using ciphertext and generated inverse key
 */
static PyObject* idea_optimized_decrypt(PyObject *self, PyObject *args)
{
	unsigned short* input;
	char* operation_mode;
	int inputlen;
	unsigned short keys[52];

	//Parse the parameter tuple
	if(!PyArg_ParseTuple(args, "s#s(HHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHHH)", &input, &inputlen, &operation_mode, &keys[0], &keys[1], &keys[2], &keys[3], &keys[4], &keys[5], &keys[6], &keys[7], &keys[8], &keys[9], &keys[10], &keys[11], &keys[12], &keys[13], &keys[14], &keys[15], &keys[16], &keys[17], &keys[18], &keys[19], &keys[20], &keys[21], &keys[22], &keys[23], &keys[24], &keys[25], &keys[26], &keys[27], &keys[28], &keys[29], &keys[30], &keys[31], &keys[32], &keys[33], &keys[34], &keys[35], &keys[36], &keys[37], &keys[38], &keys[39], &keys[40], &keys[41], &keys[42], &keys[43], &keys[44], &keys[45], &keys[46], &keys[47], &keys[48], &keys[49], &keys[50], &keys[51])){
		PyErr_SetString(PyExc_ValueError, "Input must be a string and key as 52-tuple of 16bit integers");
		return NULL;
	}

	//Check input padding, it must be perfectly padded
	if(inputlen%4){
		PyErr_SetString(PyExc_ValueError, "Input must be padded to a multiple of 64bit");
		return NULL;
	}

	//Read operations mode
	short mode;
	if(strcmp(operation_mode,"ECB") == 0){
		mode = 0;
	}else if(strcmp(operation_mode,"CBC") == 0){
		mode = 4;
	}else{
		PyErr_SetString(PyExc_ValueError, "An operation mode ECB or CBC must be provided");
		return NULL;
	}

	//Build a python string as return value
	PyObject* output = PyString_FromStringAndSize(NULL, inputlen-mode*sizeof(unsigned short));
	if(!output)
		return PyErr_NoMemory();	//Raise exception if out of memory

	//Get pointer to python string buffer
	unsigned short* blocks = (unsigned short*)PyString_AsString(output);

	//Copy input to temporary variable
	memcpy(blocks, input+mode, inputlen-mode*sizeof(unsigned short));

	//Encrypt blocks
	unsigned int i;
	for(i=0; i<(inputlen/sizeof(unsigned short)-mode); i+=4){
		process_block(blocks+i, keys);
		if(mode){
			blocks[i] ^= input[i];
			blocks[i+1] ^= input[i+1];
			blocks[i+2] ^= input[i+2];
			blocks[i+3] ^= input[i+3];
		}
	}

	//Return encrypted binary python string
	return output;
}

/* Process a block
 * By Handbook of Applied Cryptography, Algorithm 7.101
 */
void process_block(unsigned short *block, unsigned short* keys)
{
	//Temporary variables
	unsigned short t0, t1, t2;

	//Run the eight main rounds
	short i;
	for(i=0; i<8; i++){
		//A
		block[0] = mod_mul(block[0], keys[0+6*i]);
		block[3] = mod_mul(block[3], keys[3+6*i]);
		block[1] = mod_add(block[1], keys[1+6*i]);
		block[2] = mod_add(block[2], keys[2+6*i]);
		//B
		t0 = mod_mul(keys[4+6*i], block[0]^block[2]);
		t1 = mod_mul(keys[5+6*i], mod_add(t0, (block[1]^block[3])));
		t2 = mod_add(t0, t1);
		//C
		block[0] = block[0]^t1;
		block[3] = block[3]^t2;
		t0 = block[1]^t2;
		block[1] = block[2]^t1;
		block[2] = t0;
	}

	//Run the output round
	block[0] = mod_mul(block[0], keys[6*8]);
	t0 = block[1];
	block[1] = mod_add(block[2], keys[1+6*8]);
	block[2] = mod_add(t0, keys[2+6*8]);
	block[3] = mod_mul(block[3], keys[3+6*8]);
}


