#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numbertheory
import idea
import base64
import sha
import struct
import string

def generatekey(a):
	"""
		Generate new public and private key with bitlength "a", return (e,n), (d,n)
		a bitlength of key
		
		>>> (e,n), (d,n) = generatekey(256)
		>>> (e * d) % n == 1
		True
	"""
	
	q = numbertheory.random_prime(2**a)
	p = numbertheory.random_prime(2**a)
	n = q * p
	phi = (q - 1) * (p - 1)
	e = numbertheory.random_prime(phi-1)

	
	d = numbertheory.invert(e,phi)
	return (e,n), (d,n)


def sign(m,d):
	"""
		Sign the message "m" using the secret Key d 
		Return encryped Hash value
		
		>>> sign("Test Message",(7289105479147L, 23223255351017L))
		'8afc6b0e79f'
		>>> sign("Test Message",(1423683313897L, 20071841598607L))
		'11f4c3ad035b'
	"""	
	a = sha.new(m).digest()
	b = struct.unpack("B"*20,a)
	hashvalue = 0
	for i in range(0,20):
		hashvalue += (b[19-i]*(256**i)) % d[1]
	
	hashvalue = hashvalue % d[1]
	#print len(a)
	#print hashvalue
	#print hex(hashvalue)
	#print sha.new(m).hexdigest()
	
	c = hex(encrypt(hashvalue,d))
	
	#print c
	c = c[2:]
	if c[-1] == "L":
		c = c[:-1]
	return c
	 
def verify(m,c,e):
	"""
		Verify a signature compare Message m with cheksum c 
		return true if the signature is confirmed
		
		>>> c = '8afc6b0e79f'
		>>> m = "Test Message"
		>>> verify(m,c,(19342605353023L, 23223255351017L))
		2
		>>> c = '11f4c3ad035b'
		>>> m = "Test Message"
		>>> verify(m,c,(6472283450209L, 20071841598607L))
		2
	"""
	a = sha.new(m).digest()
	b = struct.unpack("B"*20,a)
	hashvalue = 0
	for i in range(0,20):
		hashvalue += (b[19-i]*(256**i)) % e[1]
	
	m = hashvalue % e[1]
	c = int(c, 16)

	return decrypt(c, e) == m

def encrypt(m,e):
	"""
		Encrypt a message "m" with the public key e
		Return an encryped message
		
		>>> encrypt(12345,(2560335680873L, 57604382841163L))
		31859130518578L
		>>> RSA.encrypt(12345,(59575615723777L, 77810448343637L))
		60484804069679L
	"""
	c = numbertheory.mod_exp(m, e[0], e[1])
	return c
	
	

def decrypt(c,d):
	"""
		Decrypt a message "c" with the privat key d
		Return the message
		
		>>>RSA.decrypt(31859130518578L,(6761834147057L, 57604382841163L))
		12345L
		>>> RSA.decrypt(60484804069679L,(33122443836733L, 77810448343637L))
		12345L
	"""

	m = numbertheory.mod_exp(c, d[0], d[1])
	return m
	
	
	

def encryptmail(m, e = None, d = None, mye = None):
	"""
		Encrypt a message "m" with IDEA and encrypt the IDEA key With RSA
		Enter a public key "e" to encrypt mail
		Enter a Private key "d" to sign mail
		
		

		My suggestion for a documentation:
----------------------------------------
def encryptmail(message, public_key = None, private_key = None, recipient_public_key = None):
	Encrypts a message if need, also reads public key if provided and verifies signature if present.
	message: To be encrypted and/or signed
	public_key: Public key of the sender, will be provided in email
	private_key: Private key of the sender, used to sign the message, if None message will not be signed
	recipient_public_key: Public key of the recipient used for encryption, if None message will not be encrypted
	Returns encrypted and/or signed message
----------------------------------------
		For examples of e-mail see decryptmail documentation
	"""
	

	mh = m
	em = m

	if mye != None or d != None or e != None:
		if mye == None:
			mye = 0,0
		m = m + "\n\nBegin Raptor Message\n\nRaptor Public Key\n" + hex(long(mye[0]))[2:-1] + "%Key%" + hex(long(mye[1]))[2:-1] + "\n\n"
		
		if d != None: #Sign Message
			cheksum = sign(mh,d)
		if e != None: #Encrypt Message with Idea
			x = numbertheory.random_number(2**128) # Idea Key
			i = idea.IDEA(x) #start Idea
			if d != None: #Sign Message
				m = m + "Raptor Signature\n" + cheksum
			c = base64.b64encode(i.encrypt(m)) #encrypt m with Idea and b64 encode the result
			RSAideakey = base64.b64encode(str(encrypt(x,e))) #encrypt Idea Key with RSA and b64 encode the result
			em = "Begin Raptor Message\n\n" + RSAideakey + "encrypt" + c
			
		if d != None and e == None: #sign no encryption
			em =  m + "Raptor Signature\n" + cheksum
		
		if d == None and e == None: # no nothing
			em = m
		
	return em

def prepare_mail(message, private_key = None, public_key = None, recipient_public_key = None):
	"""Encrypt mail with public key, sign and attach public key.
		message: Message to be prepared for transmission.
		private_key: Senders private key, message will be signed if this is provided
		public_key: Senders public key, attached if provided
		recipient_public_key: Public key of the intended recipient, used to encrypt the message
	"""
	#Add signature
	if private_key != None:
		signature = sign(message, private_key)
		message += "\r\nRaptor signature:\r\n" + signature

	#Add senders public key
	if public_key != None:
		message += "\r\nRaptor public key:\r\n"
		message += hex(long(public_key[0]))[2:-1] + "\r\n"
		message += hex(long(public_key[1]))[2:-1]

	#Encrypt message if needed
	if recipient_public_key != None:
		x = numbertheory.random_number(2**128) % recipient_public_key[1]	#IDEA key
		i = idea.IDEA(x, "CBC")		#Create instance of IDEA
		#encrypt message with Idea and base64 encode
		ciphertext = base64.b64encode(i.encrypt(message))
		#Encrypt IDEA key with RSA and base64 encode the result
		cipherkey = base64.b64encode(str(encrypt(x, recipient_public_key)))
		#Create a new message
		message = "Hi,\nThis message have been encrypted using RaptorMail you need to decrypt it using the proper private key and RaptorMail.\nRegards your guardian Raptor."
		message += "\r\nRaptor encrypted mail:\r\n"
		message += cipherkey + "\r\n"
		message += ciphertext

	return message

def process_mail(message, private_key = None):
	"""Decrypts a message if need, also reads public key if provided and verifies signature if present.
		message: Encrypted, signed and/or unsigned mail
		private_key: Private key, needed of message must be decrypted
		Returns a tuple of:
			decrypted message
			boolean True if message was decrypted
			boolean True if signature present and verified
			public_key or an empty list if public key not present

				Email suffix example:
		----------------------------------------
		\r\nRaptor encrypted mail:\r\n
		<Base-64 string>
		----------------------------------------
		\r\nRaptor signature:\r\n
		<Base-64 string>\r\n
		Raptor public key:\r\n
		<Base-64 string>
		----------------------------------------
	"""
	#Remove added CRLF if needed 
	if message[-2:] == "\r\n":
		message = message[:-2]

	#Decrypt message if encrypted
	msg_parts = message.split("\r\n")
	decrypted = False
	if len(msg_parts) >= 3 and msg_parts[-3] == "Raptor encrypted mail:":
		ideakey = msg_parts[-2]
		message = msg_parts[-1]
		if private_key == None:
			raise ValueError, "You need the private key to decrypt this message"
		#Decrypt idea key using RSA
		ideakey = decrypt(int(base64.b64decode(ideakey)), private_key)
		#Decrypt
		i = idea.IDEA(ideakey, "CBC")	#Create instance of IDEA
		try:	#Try to decrypt if this fails, the _unpad_data method ususally throws an exception
			message = i.decrypt(base64.b64decode(message))
			decrypted = True	#Mark the message as decrypted in return tuple
		except:	#If decryption fails, just use the first of the message as message, as it's an this is encrypted message
			message = "\r\n".join(msg_parts[:-2])
			decrypted = False
		#Split again
		msg_parts = message.split("\r\n")

	signature = None
	publickey = []	
	metaparts = 0		#Number of parts to remove from msg_parts to get raw message
	
	#Get signature or public key
	if len(msg_parts) >= 2 and msg_parts[-2] == "Raptor signature:":
		signature = msg_parts[-1]
		metaparts += 2
	elif len(msg_parts) >= 3 and msg_parts[-3] == "Raptor public key:":	
		publickey += [msg_parts[-2]]
		publickey += [msg_parts[-1]]
		metaparts += 3
	#Get signature or public key
	if len(msg_parts) >= (metaparts+2) and msg_parts[-(metaparts+2)] == "Raptor signature:":
		signature = msg_parts[-(metaparts+1)]
		metaparts += 2
	elif len(msg_parts) >= (metaparts+3) and msg_parts[-(metaparts+3)] == "Raptor public key:":	
		publickey += [msg_parts[-(metaparts+2)]]
		publickey += [msg_parts[-(metaparts+1)]]
		metaparts += 3

	#Parse public key
	if publickey != []:
		pubkey = [int(publickey[0], 16), int(publickey[1],16)]
	else:
		pubkey = [0,0]

	#Get raw message
	if metaparts == 0: metaparts = None
	else: metaparts = -metaparts
	message = "\r\n".join(msg_parts[0:metaparts])

	#Verify signature if present
	if pubkey != [0,0] and signature != None:
		#Verify signature
		integrity = verify(message , signature, pubkey)
	else:
		integrity = False

	return message, decrypted, integrity, pubkey
	
def decryptmail(em, e = None, d = None):
	"""
		Decrypt the IDEA key with RSA and decrypt message with IDEA and verify signature
		0: if there is no signature
		1: if there is a bad signature
		2: if there is a good signature



		My suggestion for a documentation:
----------------------------------------
def decryptmail(message, private_key = None):
	Decrypts a message if need, also reads public key if provided and verifies signature if present.
	message: Encrypted, signed and/or unsigned mail
	private_key: Private key, needed of message must be decrypted
	Returns a tuple of:
			decrypted message
			boolean True if message was decrypted
			boolean True if signature present and verified
			public_key or an empty list if public key not present
----------------------------------------

	Example of return values where message have been decrypted, signature verified, and signature is [e,n]:
"bla bla message", True, True, [e, n]
	Another example where message have not been decrypted as it was not encrypted, and signature was true and public key was [e,n]
"bla bla message", False, True, [e, n]
	Example of return values from message that's been decrypted, but not signed and no public key was provided.
"bla bla message", True, False, []


		Email suffix example:
----------------------------------------
Raptor encrypted mail:\r\n
<Base-64 string>\r\n
Raptor signature:\r\n
<Base-64 string>\r\n
Raptor public key:\r\n
<Base-64 string>
----------------------------------------

		Ordered randomly, so:
----------------------------------------
Raptor public key:\r\n
<Base-64 string>\r\n
Raptor signature:\r\n
<Base-64 string>
----------------------------------------

		Is valid too!!!

		Do this by splitting it the message:
parts = message.split("\r\n")
if parts[-2] == "Raptor signature:":
	parts[-1] is the base 64 formated signature
if parts[-4] == "Raptor signature:":
	parts[-1] is the base 64 formated signature
if parts[-6] == "Raptor signature:":
	parts[-1] is the base 64 formated signature

		Do the same with "Raptor public key:" and "Raptor encrypted mail:" notice that parts[-2] may be out of index if len(parts) < 7

try:
	Signature_verified = verify(Message.... bal...)
except:
	Signature_verified = False

		Example of another valid message:
----------------------------------------
Hi,
Now I'm wrint a message the last one said something like\r\n
Raptor public key:\r\n
111111\r\n
Raptor signature:\r\n
55555555hej meddig
------
Now I don't know why this is..\r\n
can yo tell me.\r\n
regards me.\r\n
Raptor public key:\r\n
444444444\r\n
Raptor signature:\r\n
555555555
----------------------------------------
	The above message is valid and if signature 555555555 verifies using the public key 444444444
	The two lines:
----------------------------------------
Raptor public key:\r\n
111111\r\n
----------------------------------------
	Are ignored because they are not at the end of message
	"""	
	
	"""
	print em
	print ""
	print ""
	"""
	signature = 0
	if "Begin Raptor Message" in em:
	
		tempsplit = em.split("Begin Raptor Message")
		tempsplit2 = tempsplit[len(tempsplit)-2]
		tempsplit3  = tempsplit2.split("\n\n")
		textmessage = tempsplit3[0]
		raptormessage = tempsplit[len(tempsplit)-1]
		
		if "Raptor Signature" in raptormessage: # there is signnature
			if e == None:
				raise ValueError , "No public key given to confirm the signature"
				
			tempsplit = raptormessage.split("Raptor Signature")
			
			signature = verify(textmessage,tempsplit[1],e)
			m = tempsplit[0]
		if "encrypt" in em:
			if d== None:
				raise ValueError , "No private key given to decrypt mail"
			tempsplit = raptormessage.split("encrypt")
			
			
			tempsplit2 = tempsplit[0].split("\n\n")

			Ideakey = decrypt(int(base64.b64decode(tempsplit2[1])),d)
			
			i = idea.IDEA(Ideakey) #start Idea
			m = i.decrypt(base64.b64decode(tempsplit[1]))	
			if "Raptor Signature" in m: #there is signature
				if e == None:
					raise ValueError , "No public key given to confirm the signature"
				tempsplit = m.split("Raptor Signature\n")
				
				tempsplit3 = m.split("\n\nBegin Raptor Message\n")
				m = tempsplit3[0]
				
				tempsplit3 = tempsplit[1].split("\n\n")
	
				signature = verify(m,tempsplit[1],e)
				m = tempsplit[0]
				
		if "Raptor Public Key" in m: #there public key attchments
			

			tempsplit = m.split("Raptor Public Key")
			tempsplit2 = tempsplit[0]
			if "encrypt" in em:
				tempsplit3 = tempsplit[0].split("\n\nBegin Raptor Message\n")
				m = tempsplit3[0]
			else:
				tempsplit3 = textmessage.split("\n\n")
				m = tempsplit3[0]
			attchmentsKey= tempsplit[1]
			
			tempsplit = attchmentsKey.split("%Key%")
			tempsplit2 = tempsplit[0].split("\n")
			tempsplit3 = tempsplit[1].split("\n\n")
			"""
			m = string.replace(m,"***Raptor Public Key***","%%Raptor Public Key%%")
			m = string.replace(m,"***Key***","%%%Key%%%")
			m = string.replace(m,"***Raptor Signature***","%%Raptor Signature%%")
			m = string.replace(m,"***encrypt***","%%encrypt%%")
			"""
			return m , [tempsplit2[1],tempsplit3[0]] , signature
		"""
		m = string.replace(m,"***Raptor Public Key***","%%Raptor Public Key%%")
		m = string.replace(m,"***Key***","%%%Key%%%")
		m = string.replace(m,"***Raptor Signature***","%%Raptor Signature%%")
		m = string.replace(m,"***encrypt***","%%encrypt%%")	
		"""
		return m , [0,0] , signature
	return em ,[0,0],0
def maintest():

	"""
		Complet RSA system Chek
	"""	
	keyerrors = 0
	print "Test Keygeneration"
	for i in range(0,2):
		e,d = generatekey(64)
		if RSAtest(e,d) == "FEJL i key":
			keyerrors += 1
	e,d = generatekey(256)
	e2,d2 = generatekey(256)			
	print str(keyerrors) + " Error In keygeneration"
	if keyerrors == 0:
		errors = 0
	else: 
		errors = 1
	
	print "Test Signature"
	m = "TEST OKAY"
	em = encryptmail(m,d = d)
	m = decryptmail(em,e = e)
	if m[0] == "TEST OKAY" and m[1] == 2:
		print "Signature Okay"
	else:
		errors +=1
		print "Signature Fail"
	print "Test encryption"
	m = "TEST OKAY"
	em = encryptmail(m,e = e2)
	m = decryptmail(em,d = d2)
	if m[0] == "TEST OKAY" and m[1] == 0:
		print "encryption OkaY"	
	else:
		errors +=1
		print "encryption Fail"
	print "Test sign and encryption"
	m = "TEST OKAY"
	em = encryptmail(m,e,d2)
	m = decryptmail(em,e2,d)
	if m[0] == "TEST OKAY" and m[1] == 2:
		print "Signature and encryption OkaY"		
	else:
		errors +=1
		print "Signature and encryption Fail"
	print ""
	if errors == 0:
		print "System chek complite no Errors were found"
	else:
		print "System chek complite " + errors + " Errors were found"

def test():

	"""
	import doctest
	doctest.testmod()

	if __name__ == "__main__":
		_test()
	"""
	mye = 11675129332331275246491475554834848277197103683748687788724535337905669347449L, 34394026793495475424842174728177040196789835171977895448344639014608644413403L
	myd = 3527742662381476563995979257973043342939900787894487640262082707568954261857L, 34394026793495475424842174728177040196789835171977895448344639014608644413403L
	
	fe = 623818968646347946204840299385227146880741056755614940327323541082841322451L, 3875829967294471742071526034198923123727882301613635331204833265542081995133L
	
	fd = 1080213354768690851426404283194802751603823084491954055044921675120232565083L, 3875829967294471742071526034198923123727882301613635331204833265542081995133L
	

	m = "Test tasdfasdæ knvoinroqienvori\rhis message encrypt sing"
	print "Encrypt mail"
	print "************************START OF MAIL***********************"
	c = prepare_mail(m, None, None, fe)
	print c
	print "\n***********************END OF MAIL************************"

	print process_mail(c, fd)



	print "Sign mail"
	print "************************START OF MAIL***********************"
	c = prepare_mail(m, myd, mye, None)
	print c
	print "\n***********************END OF MAIL************************"

	print process_mail(c, fd)


	print "Encrypt and sign mail"
	print "************************START OF MAIL***********************"
	c = prepare_mail(m, myd, mye, fe)
	print c
	print "\n***********************END OF MAIL************************"

	print process_mail(c, fd)

	return None

	em = encryptmail(m, mye = mye ,d = myd ,e  = fe)
	m = decryptmail(em, e = mye, d= fd)
	
	print "cifertext***************************************************"
	print em
	print "\nendcifertext**********************************************\n"
	print "Output"
	print m 
	print ""
	m = "Test this message sign"
	print m
	em = encryptmail(m, mye = mye ,d = myd )
	m = decryptmail(em, e = mye)
	print "cifertext***************************************************"
	print em
	print "\nendcifertext**********************************************\n"
	print "Output"
	print m 
	print ""
	m = "Test this message encrypt"
	print m
	em = encryptmail(m, mye = mye ,e  = fe)
	m = decryptmail(em, d= fd)
	print "cifertext***************************************************"
	print em
	print "\nendcifertext**********************************************\n"
	print "Output"
	print m 
	print ""
	m = "Test this message encrypt sign no pub"
	print m
	em = encryptmail(m, d = myd ,e = fe)
	m = decryptmail(em, e = mye, d = fd)
	print "cifertext***************************************************"
	print em
	print "\nendcifertext**********************************************\n"
	print "Output"
	print m 
	print ""
	m = "Test this message sing no pub"
	print m
	em = encryptmail(m, d = myd )
	m = decryptmail(em, e = mye)
	print "cifertext***************************************************"
	print em
	print "\nendcifertext**********************************************\n"
	print "Output"
	print m 
	print ""
	m = "Test this message encrypt no pub"
	print m
	em = encryptmail(m , e = fe)
	m = decryptmail(em, d= fd)
	print "cifertext***************************************************"
	print em
	print "\nendcifertext**********************************************\n"
	print "Output"
	print m 
	print ""
	m = "Test this message encrypt no nothing"
	print m
	em = encryptmail(m )
	m = decryptmail(em)
	print "cifertext***************************************************"
	print em
	print "\nendcifertext**********************************************\n"
	print "Output"
	print m 
	print ""
	
	return "end"
	
def RSAtest(e,d):
	"""
		TEST the RSA key
	"""	
	m = 55
	c = numbertheory.mod_exp(m, e[0], e[1])
	m = numbertheory.mod_exp(c, d[0], d[1])

	if m == 55:
		return "det Virker"
	return "FEJL i key"


#TODO: Stub out methods to serialize keys and messages
#TODO: Do something like padding...
