#!/usr/bin/env python
# -*- coding: utf-8 -*-

import imaplib
import string
import StringIO
import pickle
import email
import email.header
import imap_parser
import threading
import smtplib
from pysqlite2 import dbapi2 as sqlite3
#SQLite notes: /usr/share/doc/python-pysqlite2/examples
#pysqlite.org

class GMail:
	def __init__(self, username, password, dbpath = ":memory:"):
		"""	Initialize GMail instance
			username: username to login with, user must have imap enabled
			password: password associated with the user
			dbpath: Path to the database, if not specified database will be stored in memory
		"""
		#Check types
		if not isinstance(username, basestring):
			raise ValueError, "Username is not a string"
		if not isinstance(password, basestring):
			raise ValueError, "Password is not a string"
		self.username = username
		self.password = password
		if not isinstance(dbpath, basestring):
			raise ValueError, "dbpath is not a string"

		#Create imap connection lazily
		self.__imap = None

		#Load database
		self._dbpath = dbpath
		self._dbconn = sqlite3.connect(dbpath)	#TODO: Exception handling
		self.database = self._dbconn.cursor()	#Public database cursor
		self._database_thread = threading.currentThread().getName()
		self.database.execute("SELECT count(name) from sqlite_master where type='table'")
		if self.database.fetchone()[0] < 6:
			self._createtables(self.database)
			self._dbconn.commit()

		#Setup possible events
		self.progress_report = None		#takes an argument of progress, 0-100, once complete 100 is parsed, not threadsafe

		#Internal progress state
		self.progress = float(0)

	def _login(self, username = None, password = None):
		"""Ensures that __imap is a connection to gmail"""
		#Use new provided username and password
		if username and password:
			self.username = username
			self.password = password

		#Create connection lazily
		if not self.__imap:
			self.__imap = imaplib.IMAP4_SSL("imap.gmail.com", 993)

		#Test the connection using noop
		try:
			st, [res] = self.__imap.noop()
			if res != "Success":
				self.__imap.login(self.username, self.password)
		except imaplib.IMAP4.abort:
			try:
				self.__imap = imaplib.IMAP4_SSL("imap.gmail.com", 993)
				self.__imap.login(self.username, self.password)
			except:
				return False
		except imaplib.IMAP4.error:
			try:
				self.__imap.login(self.username, self.password)
				self.__imap.noop()
			except:
				try:
					self.__imap = imaplib.IMAP4_SSL("imap.gmail.com", 993)
					self.__imap.login(self.username, self.password)
				except:
					return False
		return True

	def get_database(self):
		"""Gets tuple with database connection and cursor
			Creates a new one if not on mainthread"""
		#Check if we need a new connection
		if threading.currentThread().getName() != self._database_thread:
			conn = sqlite3.connect(self._dbpath)
			cur = conn.cursor()
			return conn, cur
		else:
			return self._dbconn, self.database

	def return_database(self, connection, cursor):
		"""Commit and close database connection and cursor if needed"""
		connection.commit()
		if cursor != self.database:
			cursor.close()
		if connection != self._dbconn:
			connection.close()

	def sync_labels(self, progress_complete = 100, cur = None):
		"""Synchronizes labels
			progress_complete: Amount of progress to have added to self.progress when completed
			cur: database cursor for this thread
			Returns a list of current labels
		"""
		#Make sure we're logged in and have connection
		if not self._login():
			raise ValueError, "Couldn't login at GMail ensure that you have internet access and correct password/username"

		#Get a connection if needed
		if cur == None:
			conn, cur = self.get_database()

		#Report progress
		self.progress += float(progress_complete)/2
		if self.progress_report: self.progress_report(self.progress)

		#List labels and add them to labels if not already there
		st, raw = self.__imap.list()
		if st != "OK":
			pass	#Bad thing happend
		labels = []
		for data in raw:
			attributes, root, name = self._parse(data)
			if "\Noselect" in attributes:
				continue	#These are not labels
			if name in ["[Gmail]", "[Gmail]/Spam", "[Gmail]/Trash"]:
				continue	#Don't index dynamic labels
			self._addlabeltodatabase(name, cur)
			labels += [name]

		#Delete non-existing labels from database
		cur.execute("SELECT name FROM labels")
		data = cur.fetchall()
		for row in data:
			if row[0] not in labels:
				self._deletelabelfromdatabase(row[0], cur)

		#Report progress
		self.progress += float(progress_complete)/2
		if self.progress_report: self.progress_report(self.progress)

		#Return db if needed
		if locals().has_key('conn'):
			self.return_database(conn, cur)

		#Return labels
		return labels

	def sync_actions(self, progress_complete = 100, cur = None, labels = None):
		"""Execute actions on remote server
			progress_complete: Amount of progress to have added to self.progress when completed
			cur: database cursor for this thread
			labels: A list of all labels in database
		"""
		#Make sure we're logged in and have connection
		if not self._login():
			raise ValueError, "Couldn't login at GMail ensure that you have internet access and correct password/username"
		#Get a connection if needed
		if cur == None:
			conn, cur = self.get_database()
		#Get labels if we don't have them
		if labels == None:
			labels = self.get_labels(cur = cur)

		#Report progress
		self.progress += float(progress_complete)/5
		if self.progress_report: self.progress_report(self.progress)

		#Execute delete actions
		self.__imap.select("[Gmail]/All Mail")
		cur.execute("SELECT msg_id FROM delete_actions")
		msg_ids = cur.fetchall()
		for [msg_id] in msg_ids:
			#Get the uid or id of this message by searching all mail
			st, [data] = self.__imap.search(None, '(HEADER MESSAGE-ID "' + msg_id + '")')
			#Ignore errors in getting the message as it may have been deleted in webclient
			if st == "OK" and " " not in data and data != "":
				self.__imap.copy(data, "[Gmail]/Trash")
				self.__imap.store(data, "+FLAGS", "\\Deleted")
				self._delete_message_from_database(msg_id, cur)
		if msg_ids:
			self.__imap.expunge()
		#Delete executed actions
		cur.execute("DELETE FROM delete_actions")

		#Report progress
		self.progress += float(progress_complete)/5
		if self.progress_report: self.progress_report(self.progress)

		#Add labels
		cur.execute("SELECT name, msg_id FROM label_actions WHERE action = 'add'")
		new_labels_associations = cur.fetchall()
		for [label, msg_id] in new_labels_associations:
			if label not in labels:
				try:
					st, data = self.__imap.create(label)	#TODO: Exception handling as some label names are really bad!!!
					if st == "BAD":
						pass #We should notice this!
					#Add label to database as it may have been deleted by sync_labels
					label_rowid = self._addlabeltodatabase(label, cur)
					cur.execute("SELECT COUNT(name) FROM label_relations WHERE name = ? and msg_id = ?", [label_rowid, msg_id])
					[labelassoc] = cur.fetchone()
					if labelassoc == 0:	#Re associate label if association deleted by sync_labels
						cur.execute("INSERT INTO label_relations (label_rowid, msg_id) VALUES (?, ?)", [label_rowid, msg_id])
				except:
					pass	#Don't care about badnames then label only exists locally
			st, [data] = self.__imap.search(None, '(HEADER MESSAGE-ID "' + msg_id + '")')
			#Ignore errors in getting the message as it may have been deleted in webclient
			if st == "OK" and " " not in data and data != "":
				self.__imap.copy(data, label)
		#Delete executed actions
		cur.execute("DELETE FROM label_actions WHERE action = 'add'")

		#Report progress
		self.progress += float(progress_complete)/5
		if self.progress_report: self.progress_report(self.progress)

		#Remove labels
		cur.execute("SELECT name, msg_id FROM label_actions WHERE action = 'remove'")
		removed_labels_associations = cur.fetchall()
		for [label, msg_id] in removed_labels_associations:
			st, [data] = self.__imap.select(label)
			if st == "OK":
				st, [data] = self.__imap.search(None, '(HEADER MESSAGE-ID "' + msg_id + '")')
				if st == "OK" and " " not in data and data != "":
					self.__imap.store(data, "+FLAGS", "\\Deleted")
		if removed_labels_associations:
			self.__imap.expunge()
		#Delete executed actions
		cur.execute("DELETE FROM label_actions WHERE action = 'remove'")

		#Report progress
		self.progress += float(progress_complete)/5
		if self.progress_report: self.progress_report(self.progress)

		#Update the read state of messages at the server
		cur.execute("SELECT msg_id, read FROM mark_read_actions")
		self.__imap.select("[Gmail]/All Mail")
		for [msg_id, read] in cur:
			st, [data] = self.__imap.search(None, '(HEADER MESSAGE-ID "' + msg_id + '")')
			if st == "OK" and data != "":
				if read:
					self.__imap.store(data, "+FLAGS", "\\Seen")
				else:
					self.__imap.store(data, "-FLAGS", "\\Seen")
		#Delete executed actions
		cur.execute("DELETE FROM mark_read_actions")

		#Report progress
		self.progress += float(progress_complete)/5
		if self.progress_report: self.progress_report(self.progress)

		#Return db if needed
		if locals().has_key('conn'):
			self.return_database(conn, cur)

	def sync_mails(self, label = None, progress_complete = 100, cur = None, labels = None):
		"""Synchronizes messages from a label
			progress_complete: Amount of progress to have added to self.progress when completed
			label: Label to synchronize, if None all will be synced
			cur: database cursor for this thread
			Returns a list of msg_id's of the new emails
		"""
		#Make sure we're logged in and have connection
		if not self._login():
			raise ValueError, "Couldn't login at GMail ensure that you have internet access and correct password/username"
		#Get a connection if needed
		if cur == None:
			conn, cur = self.get_database()
		#Get labels if we don't have them
		if labels == None:
			labels = self.get_labels(cur = cur)

		#Create list of new email msg_id's
		new_mails = []

		#If we must sync all labels, loop through labels and call this method
		if label == None:
			progress_per_label = float(progress_complete)/len(labels)
			for label in labels:
				new_mails += self.sync_mails(label, progress_per_label, cur, labels)
				if locals().has_key('conn'): conn.commit()
		else:
			#If we have a given label to sync, then lets do just that
			self.__imap.select(label)
			cur.execute("SELECT last_uid, rowid FROM labels WHERE name=? LIMIT 1", [label])
			[last_uid, label_rowid] = cur.fetchone()
			st, raw = self.__imap.uid("fetch", last_uid + ":*", "(BODY[HEADER.FIELDS (Message-ID)] FLAGS)")
			if st != "OK":
				pass	#Bad thing happend

			#Setting up progress
			progress_per_mail = float(progress_complete) / len(raw)

			#Looping thought the list of mails
			for dataset in raw:
				#Report progress
				self.progress += progress_per_mail
				if self.progress_report: self.progress_report(self.progress)

				#Skip those weird ')' chars that python devs for some reason decided to not remove from IMAP output
				if not isinstance(dataset, tuple) or dataset == ")":
					continue	#Ignore the closing ')' stings
				#Get UID and msg_id
				data = self._parse(dataset[0] + ")")
				UID = data[1][1]
				flags = data[1][3]
				read = "\\Seen" in flags	#Check if the message is read
				data = email.message_from_string(dataset[1])
				msg_id = data["Message-ID"]

				#Check if we have the message
				cur.execute("SELECT read FROM mails WHERE msg_id=?", [msg_id])
				readstate = cur.fetchone()
				if readstate == None:
					#Fetch mail
					st, rawmail = self.__imap.uid("fetch", UID, "(RFC822)")
					if st != "OK":
						pass #Bad thing happend

					#Parse the mail
					mail = email.message_from_string(rawmail[0][1])

					#Serialize the mail
					buf = StringIO.StringIO()
					pickle.dump(mail, buf, -1)
					message = sqlite3.Binary(buf.getvalue())

					#Extract metedata and decode it
					[sender,		sender_enc]		= email.header.decode_header(mail["From"]) [0]
					[recipient,	recipient_enc]		= email.header.decode_header(mail["To"]) [0]
					[subject,		subject_enc]		= email.header.decode_header(mail["Subject"]) [0]
					if sender_enc:
						sender = sender.decode(sender_enc)
					sender = unicode(sender)
					if recipient_enc:
						recipient = recipient.decode(recipient_enc)
					recipient = unicode(recipient)
					if subject_enc:
						subject = subject.decode(subject_enc)
					subject = unicode(subject)
					date = email.utils.mktime_tz(email.utils.parsedate_tz(mail["date"]))

					#Get relations
					ref_ids = []
					if mail["References"]:
						for ref_id in mail["References"].split("\r\n\t "):
							ref_ids += [ref_id]	#Check if it contains an AT character since, gmail add random stuff that doesn't
					if mail["In-Reply-To"]:
						for ref_id in mail["In-Reply-To"].split("\r\n\t "):
							if ref_id not in ref_ids: ref_ids += [ref_id]
					relations = ",".join(ref_ids)
					"""
					Just in case you want to implement search support

text = mail.get_payload(decode=True)
enc = mail.get_content_charset()
if enc != None:
	text = text.decode(enc)


Get text index of an HTML message
from BeautifulSoup import BeautifulSoup 
b = BeautifulSoup(tt)
"".join(b.findAll(text=True))

.replace("\n", " ").replace("\r", " ").replace("\t"," ").replace("  "," ").replace("  "," ").replace("  "," ")
					"""
					#Add to database
					values = [msg_id, read, sender, recipient, subject, relations, date, message]
					cur.execute("INSERT INTO mails (msg_id, read, sender, recipient, subject, relations, date, message) VALUES (?,?,?,?,?,?,?,?)", values)
					
					#Add to a list of new mail
					new_mails += [msg_id]
				#Update read field if needed, don't happend if message is new
				elif read == readstate[0]:
					self._set_read_state(msg_id, read, cur)

				#Add label relation to message
				self._addlabelrelationtodatabase(msg_id, labelrowid = label_rowid, cur = cur)

				#Update last_uid
				if int(UID) > int(last_uid):
					#Update last_uid for every 10 mails, so that if connection dies, we wont start from scratch
					if (int(UID) - int(last_uid)) % 10 == 9:
						cur.execute("UPDATE labels SET last_uid=? WHERE name=?", [UID, label])
					#Update last_uid
					last_uid = UID

			#Save last_uid to database, so we don't start from scratch next time
			cur.execute("UPDATE labels SET last_uid=? WHERE name=?", [last_uid, label])

		#Return db if needed
		if locals().has_key('conn'):
			self.return_database(conn, cur)

		#Return list of new mails
		return new_mails

	def sync(self, cur = None):
		"""	Synchronizes client and server
			cur: database cursor for this thread
			Returns a list of new mails

			This method performs the following actions
			 - Execute stored actions as discussed in draft-ietf-imap-disc
			 - Add new labels from server
			 - Delete labels no longer on server
			 - Fetch Message-ID for all labels from last known UID to *
			   - Fetch mail and store it if not present in database
			   - Add label association if not present in database
			Remarks:
			Labels [Gmail], [Gmail]/Spam, [Gmail]/Trash or the \Noselect attribute are not
			 Synchronized, as these are consider dynamic. Support for listing them may be
			 added later.
			Mails that have been deleted are not detected, this means that you'll not be
			 able to delete message on the server and expect them to disappear on the client.
			 This may have consequenses for the [Gmail]/Drafts label as UIDs may be updated
			 every time a draft is saved, if this is not the case changes to drafts are not
			 detected. Workaround may be found later.
		"""
		#Make sure we can login
		if not self._login():
			raise ValueError, "Couldn't login at GMail ensure that you have internet access and correct password/username"

		#Set progress 0
		self.progress = 0
		if self.progress_report: self.progress_report(self.progress)

		#Get a connection if needed
		if cur == None:
			conn, cur = self.get_database()

		labels = self.sync_labels(5, cur)
		self.sync_actions(10, cur, labels)
		new_mails = self.sync_mails(None, 75, cur, labels)

		#Update progress
		if self.progress_report: self.progress_report(100)

		#Return db if needed
		if locals().has_key('conn'):
			self.return_database(conn, cur)
		
		#Return list of new emails
		return new_mails

	def _set_read_state(self, msg_id, read = True, cur = None):
		"""Marks a message as read/unread in database
			msg_id: Message-ID of message to mark
			read: True/False as Read/Unread respectively
			cur: database cursor for this thread
		"""
		#Check parameters
		if not isinstance(msg_id, basestring):
			raise ValueError, "msg_id must be a string"
		if not isinstance(read, bool):
			raise ValueError, "read must be a boolean True/False"

		#Get a connection if needed
		if cur == None:
			conn, cur = self.get_database()

		cur.execute("UPDATE mails SET read = ? WHERE msg_id = ?", [read, msg_id])

		#Return db if needed
		if locals().has_key('conn'):
			self.return_database(conn, cur)

	def _addlabeltodatabase(self, labelname, cur = None):
		"""	Checks if labelname is in database and adds it if not
			labelname: label to add to database
			cur: database cursor for this thread
			Returns label_rowid of the label entry in the labels table
		"""
		#Get a connection if needed
		if cur == None:
			conn, cur = self.get_database()

		cur.execute("SELECT COUNT(name), rowid FROM labels WHERE name=? LIMIT 1", [labelname])
		result = cur.fetchone()
		#If label is not in database
		if result[0] < 1:
			cur.execute("INSERT INTO labels (name, last_uid) VALUES (?, '1')", [labelname])
			R = cur.lastrowid
		else:
			R = result[1]

		#Return db if needed
		if locals().has_key('conn'):
			self.return_database(conn, cur)

		return R

	def _deletelabelfromdatabase(self, labelname, cur = None):
		"""Deletes a label from database
			labelname: label to remove from database
			cur: database cursor for current thread
		"""
		#Get a connection if needed
		if cur == None:
			conn, cur = self.get_database()

		cur.execute("SELECT rowid FROM labels WHERE name = ? LIMIT 1", [labelname])
		labelid = cur.fetchone()[0]
		cur.execute("DELETE FROM label_relations WHERE label_rowid = ?", [labelid]);
		cur.execute("DELETE FROM labels WHERE rowid = ?", [labelid]);

		#Return db if needed
		if locals().has_key('conn'):
			self.return_database(conn, cur)

	#TODO: Search and replace these names with some that have underscor as space between words
	def _addlabelrelationtodatabase(self, msg_id, labelname = "", labelrowid = None, cur = None):
		"""	Adds a label association to the label_relations table
			msg_id: Message-ID of the message to be labeled
			labelname: Name of the label to be applied
			labelrowid: Rowid of the label to be applied
			cur: database cursor for current thread
			Remarks: Either labelname or labelrowid must be provided		
			Returns: True if added, false if it was already there
		"""
		#If we haven't got label_rowid go get it, and create the label if needed
		if labelrowid == None and label != "":
			labelrowid = self._addlabeltodatabase(labelname, cur)
		#Without label_rowid or labelname we've got an error
		elif labelrowid == None and label == "":
			raise ValueError, "Bad parameters, either labelname or label_rowid must be set"

		#Get a connection if needed
		if cur == None:
			conn, cur = self.get_database()

		#Check if label association exists
		cur.execute("SELECT COUNT(msg_id) FROM label_relations WHERE msg_id=? AND label_rowid = ?", [msg_id, labelrowid])
		#Add it of not
		if cur.fetchone()[0] < 1:
			cur.execute("INSERT INTO label_relations (label_rowid, msg_id) VALUES(?,?)", [labelrowid, msg_id])

		#Return db if needed
		if locals().has_key('conn'):
			self.return_database(conn, cur)

	def _delete_message_from_database(self, msg_id, cur = None):
		"""Deletes a message from the local database"""
		#Get a connection if needed
		if cur == None:
			conn, cur = self.get_database()

		cur.execute("DELETE FROM label_relations WHERE msg_id = ?", [msg_id])
		cur.execute("DELETE FROM mails WHERE msg_id = ?", [msg_id])

		#Return db if needed
		if locals().has_key('conn'):
			self.return_database(conn, cur)

	def _createtables(self, cur = None):
		"""	Creates Following tables:
			labels:							Labels in our database
				- rowid						id of label, automatically created
				- name			TEXT		label name
				- last_uid		TEXT		Greates UID from sync
			label_relations					Relations between messages and labels
				- label_rowid	INT			label row id
				- msg_id		TEXT		Message-ID
			mails							Mails in our database
				- msg_id		TEXT		Message-ID
				- read			INT			0 if unread, 1 if read
				- sender		TEXT		From
				- recipient		TEXT		To
				- subject		TEXT		Subject
				- relations		TEXT		References and In-Reply-To
				- date			INT			date as unix timestamp
				- message		(blob)		email.message instance
			label_actions					Labels to be removed/added at server
				- name			TEXT		label name
				- action		TEXT		"apply"/"remove"
				- msg_id		TEXT		Message-ID
			delete_actions					Messages to be deleted at server
				- msg_id		TEXT		Message-ID
			mark_read_actions				Messages to be marked read/unread at server
				- msg_id		TEXT		Message-ID
				- read			INT			0 if unread, 1 if read
		"""
		#Get a connection if needed
		if cur == None:
			conn, cur = self.get_database()

		cur.executescript("""
			CREATE TABLE labels(
				name		TEXT,
				last_uid	TEXT
			);
			CREATE TABLE label_relations(
				label_rowid	INT,
				msg_id		TEXT
			);
			CREATE TABLE mails(
				msg_id		TEXT,
				read		INT,
				sender		TEXT,
				recipient	TEXT,
				subject		TEXT,
				relations	TEXT,
				date		INT,
				message		BLOB
			);
			CREATE TABLE label_actions(
				name		TEXT,
				action		TEXT,
				msg_id		TEXT
			);
			CREATE TABLE delete_actions(
				msg_id		TEXT
			);
			CREATE TABLE mark_read_actions(
				msg_id		TEXT,
				read		INT
			);
		""")

		#Return db if needed
		if locals().has_key('conn'):
			self.return_database(conn, cur)

	def _extract_content(self, message, mimetype = "text/plain"):
		"""Extract and decode the content from a message
			message: instance of email.message.Message to decode
			mimetype: The wanted mimetype, should only be text/html or text/plain, defaults to the latter
			Returns a tuple of payload and mimetype
		"""
		#Check parameters
		if not isinstance(message, email.message.Message):
			raise ValueError, "Message is not an instance of email.message.Message"
		if mimetype not in ["text/plain", "text/html"]:
			raise ValueError, "This method only handles text/plain and text/html"

		#Get parts from the message
		html_part = None
		text_part = None
		for part in message.walk():
			if part.get_content_type() == "text/plain":
				text_part = part
			elif part.get_content_type() == "text/html":
				html_part = part

		#Find the part we'll decode
		if html_part == None and text_part == None:
			return False
		if (mimetype == "text/plain" and text_part != None) or (mimetype == "text/html" and html_part == None):
			part = text_part
		elif (mimetype == "text/html" and html_part != None) or (mimetype == "text/plain" and text_part == None):
			part = html_part

		#Decode the part
		text = part.get_payload(decode=True)
		encoding = part.get_content_charset()
		mimetype = part.get_content_type()
		if encoding != None:
			text = text.decode(encoding)

		#Return text and mimetype
		return text, mimetype

	def __extract_content_from_messages(self, messages, mimetype = "text/plain", colindex = -1):
		"""Extract and decode content from a list of messages, or table containing a column of messages
			messages: List or table of messages as pickled strings (table as a list of lists)
			mimetype: The preferred mimetype to return
			colindex: -1 if messages is a list of messages, otherwise it indicates the index of the messages
					   in the sublists of messages
		"""
		#Check parameters
		if mimetype not in ["text/plain", "text/html"]:
			raise ValueError, "This method only handles text/plain and text/html"
		if not isinstance(messages, (list, tuple)):
			raise ValueError, "messages must be a list, otherwise use the _extract_content method"
		#Create 
		result = []
		for msg in messages:
			#Get raw message
			if isinstance(msg, (list, tuple)):
				raw_msg = msg[colindex]
			else:
				raw_msg = msg
			#Decode raw message
			email_msg = pickle.loads(raw_msg)
			content, content_mime = self._extract_content(email_msg, mimetype)
			#Add to return list
			if isinstance(msg, (list, tuple)):
				toadd = list(msg)
				toadd[colindex] = [content, content_mime]
				result += toadd
			else:
				result += [content, content_mime]
		return result
	
	def _parse(self, data):
		"""Parse a parenthesized list as defined by RFC 3501 section 4.4"""
		return imap_parser.recursive_parser(data)	

	def add_label(self, msg_id, labelname):
		"""Add label to a message
			msg_id: Id of the message to add a label to
			labelname: name of the label to add

			Remarks: labelname can also be special names such as INBOX
		"""
		#Check parameters
		if not isinstance(msg_id, basestring):
			raise ValueError, "msg_id must be a string"
		if not isinstance(labelname, basestring):
			raise ValueError, "labelname must be a string"

		#Remove bad chars from label name
		labelname = labelname.replace('\"', '')
		labelname = labelname.replace("'", '')
		labelname = labelname.replace(' ', '_')

		#Get database
		conn, cur = self.get_database()

		#Create label if doesn't exist
		rowid = self._addlabeltodatabase(labelname, cur)

		#Check if a label_relations already exists
		cur.execute("SELECT msg_id FROM label_relations WHERE msg_id = ? AND label_rowid = ?", [msg_id, rowid])
		check = cur.fetchone()
		if check == None:
			#Check if it's in label_actions
			cur.execute("SELECT action FROM label_actions WHERE name = ? AND msg_id = ?", [labelname, msg_id])
			action = cur.fetchone()
			if action == None:
				cur.execute("INSERT INTO label_actions (name, action, msg_id) VALUES(?, 'add', ?)", [labelname, msg_id])
			elif action != "add":
				cur.execute("UPDATE label_actions SET action = 'add' WHERE msg_id = ? AND name = ?", [msg_id, labelname])
			#Add a label relation
			cur.execute("INSERT INTO label_relations (msg_id, label_rowid) VALUES (?, ?)", [msg_id, rowid])

		#Return the database
		self.return_database(conn, cur)

	def remove_label(self, msg_id, labelname):
		"""Remove label from a message
			msg_id: Id of the message to remove a label from
			labelname: name of the label from remove from

			Remarks: labelname can also be special names such as INBOX
		"""
		#Check parameters
		if not isinstance(msg_id, basestring):
			raise ValueError, "msg_id must be a string"
		if not isinstance(labelname, basestring):
			raise ValueError, "labelname must be a string"

		#Get database
		conn, cur = self.get_database()

		#Verify that a label_relations exists
		cur.execute("SELECT msg_id, label_rowid FROM label_relations WHERE msg_id = ? AND label_rowid = (SELECT rowid FROM labels WHERE name = ?)", [msg_id, labelname])
		data = cur.fetchone()
		if data != None:
			#Unpack data
			[msg_id, label_rowid] = data

			#Check if it's in label_actions
			cur.execute("SELECT action FROM label_actions WHERE name = ? AND msg_id = ?", [labelname, msg_id])
			action = cur.fetchone()
			if action == None:
				cur.execute("INSERT INTO label_actions (name, action, msg_id) VALUES(?, 'remove', ?)", [labelname, msg_id])
			elif action != "remove":
				cur.execute("UPDATE label_actions SET action = 'remove' WHERE msg_id = ? AND name = ?", [msg_id, labelname])
			#Remove label relation
			cur.execute("DELETE FROM label_relations WHERE msg_id = ? AND label_rowid = ?", [msg_id, label_rowid])
		
		#Return the database
		self.return_database(conn, cur)
		
		if data == None:
			raise ValueError, "Label cannot be removed, it's not associated with the message"

	def delete_message(self, msg_id):
		"""Delete a message"""
		#Check parameters
		if not isinstance(msg_id, basestring):
			raise ValueError, "msg_id must be a string"

		#Get database
		conn, cur = self.get_database()

		#Add delete action if not in database
		cur.execute("SELECT COUNT(msg_id) FROM delete_actions WHERE msg_id = ?", [msg_id])
		[count] = cur.fetchone()
		if count == 0:
			cur.execute("DELETE FROM mails WHERE msg_id = ?", [msg_id])
			cur.execute("DELETE FROM label_relations WHERE msg_id = ?", [msg_id])
			cur.execute("INSERT INTO delete_actions (msg_id) VALUES (?)", [msg_id])

		#Return the database
		self.return_database(conn, cur)

	def get_labels(self, msg_id = None, cur = None):
		"""Get list of labels associated with a message, or all labels
			msg_id: message to get labels of, if None a list of existing labels is returned
			cur: database cursor for current thread
		"""
		#Get a connection if needed
		if cur == None:
			conn, cur = self.get_database()

		#Select label names
		variables = []
		sql = "SELECT name FROM labels"

		#Condition depending on parameters
		if isinstance(msg_id, basestring):
			variables += [msg_id]
			sql += " WHERE rowid IN (SELECT label_rowid FROM label_relations WHERE msg_id = ?)"

		#Add ordering
		sql += " ORDER BY name"

		#Execute and return
		cur.execute(sql, variables)
		R = []
		for [label] in cur.fetchall():
			R += [label]

		#Return db if needed
		if locals().has_key('conn'):
			self.return_database(conn, cur)

		#Return result
		return R

	def list_label(self, label, fields = ["sender","subject","msg_id"], limit = 0, page = -1, extract = True, mimetype = "text/plain"):
		"""Returns list of message associated with a label ordered by date
			label: name of the label to list
			fields: subject, sender, recipient, msg_id, date, message, read or list thereof
			limit: Maximum number of entries to return, defaults to 0 for unlimited
			page: Pagenumber to return, -1 for all
			extract: Extract and decode content and mimetype from message, the message entry will be a list of content followed by mimetype
			mimetype: Preferred mimetype if asked to extract, text/plain and text/html supported
		"""
		#Check parameter validity
		if not isinstance(label, basestring) or len(label) == 0:
			raise ValueError, "label must be a string and longer than 0"
		if isinstance(fields, basestring):
			fields = [fields]
		elif not isinstance(fields, list):
			raise ValueError, "fields must be a string or list of strings"
		elif len(fields) == 0:
			raise ValueError, "You must request some fields in return"
		for field in fields:
			if field not in ["subject", "sender", "recipient", "msg_id", "date", "message", "read"]:
				raise ValueError, "fields may only be subject, sender, recipient, msg_id, message or list thereof"
		if not isinstance(limit, int) or limit < 0:
			raise ValueError, "limit must be an positive integer"
		if not isinstance(page, int):
			raise ValueError, "page must be an integer"
		if limit == 0 and page != -1:
			raise ValueError, "page and limit are not consistent"
		if not isinstance(extract, bool):
			raise ValueError, "extract is not boolean, it must be true or false as it'll be intepret as boolean"

		#Prepare sql command
		variables = [label]
		sql = "SELECT " + ",".join(fields) + " FROM mails WHERE msg_id IN (SELECT msg_id FROM label_relations WHERE label_rowid = (SELECT rowid FROM labels WHERE name = ?)) ORDER BY date DESC"

		#Extent sql command to support page and limit
		if limit > 0:
			sql += " LIMIT ?"
			variables += [limit]
		if page >= 0:
			sql += " OFFSET ?"
			variables += [page*limit]

		#Execute sql command
		conn, cur = self.get_database()
		cur.execute(sql, variables)
		R = cur.fetchall()

		self.return_database(conn, cur)

		#Extract content and mimetype if needed
		if extract:
			if "message" in fields:
				R = self.__extract_content_from_messages(R, mimetype, fields.index("message"))

		#If we only have one field convert R to a list instead of a list of lists
		if len(fields) == 1:
			Re = R
			R = []
			for [i] in Re:
				R += [i]

		#Return the result
		return R

	def get_conversation(self, msg_id, fields = ["sender","subject","msg_id"], limit = 0, page = -1, extract = True, mimetype = "text/plain"):
		"""Returns a list of messages in a conversation, ordered by date
			msg_id: Message-ID of a message in the conversation
			fields: subject, sender, recipient, msg_id, date, message, read or list thereof
			limit: Maximum number of entries to return, defaults to 0 for unlimited
			page: Pagenumber to return, -1 for all
			extract: Extract and decode content and mimetype from message, the message entry will be a list of content followed by mimetype
			mimetype: Preferred mimetype if asked to extract, text/plain and text/html supported
		"""
		#Check parameter validity
		if not isinstance(msg_id, basestring) or len(msg_id) == 0:
			raise ValueError, "msg_id must be a string and longer than 0"
		if isinstance(fields, basestring):
			fields = [fields]
		elif not isinstance(fields, list):
			raise ValueError, "fields must be a string or list of strings"
		elif len(fields) == 0:
			raise ValueError, "You must request some fields in return"
		for field in fields:
			if field not in ["subject", "sender", "recipient", "msg_id", "date", "message", "read"]:
				raise ValueError, "fields may only be subject, sender, recipient, msg_id, message or list thereof"
		if not isinstance(limit, int) or limit < 0:
			raise ValueError, "limit must be an positive integer"
		if not isinstance(page, int):
			raise ValueError, "page must be an integer"
		if limit == 0 and page != -1:
			raise ValueError, "page and limit are not consistent"
		if not isinstance(extract, bool):
			raise ValueError, "extract is not boolean, it must be true or false as it'll be intepret as boolean"

		#Get database connection
		conn, cur = self.get_database()

		#Create a list of related msg_ids, and add related mails for each entry, and search added entries as they appear
		msg_ids = [msg_id]
		for msg_id in msg_ids:
			cur.execute("SELECT msg_id, relations, subject FROM mails WHERE relations LIKE ? OR msg_id = ?", ["%"+msg_id+"%", msg_id])
			for relations in cur:
				if relations[0] not in msg_ids:
					msg_ids += [relations[0]]
				#Skip relations of there are none
				if len(relations[1]) != 0:
					for relation in relations[1].split(","):
						if relation not in msg_ids:
							msg_ids += [relation]

		#Prepare sql command
		variables = []
		sql = "SELECT " + ",".join(fields) + " FROM mails WHERE msg_id IN ('" + "','".join(msg_ids) + "') ORDER BY date DESC"

		#Extent sql command to support page and limit
		if limit > 0:
			sql += " LIMIT ?"
			variables += [limit]
		if page >= 0:
			sql += " OFFSET ?"
			variables += [page*limit]

		#Execute sql command
		cur.execute(sql, variables)
		R = cur.fetchall()
		self.return_database(conn, cur)

		#Extract content and mimetype if needed
		if extract:
			if "message" in fields:
				R = self.__extract_content_from_messages(R, mimetype, fields.index("message"))

		#If we only have one field convert R to a list instead of a list of lists
		if len(fields) == 1:
			Re = R
			R = []
			for [i] in Re:
				R += [i]

		#Return the result
		return R

	def get_message(self, msg_id, fields = ["msg_id", "sender", "subject" ,"message"], extract = True, mimetype = "text/plain"):
		"""Get a message using its Message-ID
			msg_id: Message-ID of the message to get
			fields: subject, sender, recipient, msg_id, date, message, read or list thereof
		"""
		#Check parameters
		if not isinstance(msg_id, basestring) or len(msg_id) == 0:
			raise ValueError, "msg_id must be a non-empty string"
		if isinstance(fields, basestring):
			fields = [fields]
		elif not isinstance(fields, list):
			raise ValueError, "fields must be a string or list of strings"
		elif len(fields) == 0:
			raise ValueError, "You must request some fields in return"
		for field in fields:
			if field not in ["subject", "sender", "recipient", "msg_id", "date", "message", "read"]:
				raise ValueError, "fields may only be subject, sender, recipient, msg_id, message or list thereof"
		if not isinstance(extract, bool):
			raise ValueError, "extract is not boolean, it must be true or false as it'll be intepret as boolean"

		#Get database
		conn, cur = self.get_database()

		#Prepare sql
		variables = [msg_id]
		sql = "SELECT " + ",".join(fields) + " FROM mails WHERE msg_id = ? LIMIT 1"

		#Execute sql
		cur.execute(sql, variables)
		R = cur.fetchall()

		#Return database
		self.return_database(conn, cur)

		#Extract if needed
		if extract:
			if "message" in fields:
				R = self.__extract_content_from_messages(R, mimetype, fields.index("message"))

		return R

	def mark_read(self, msg_id, read = True):
		"""Marks a message as read/unread in database
			msg_id: Message-ID of message to mark
			read: True/False as Read/Unread respectively
		"""
		#Check parameters
		if not isinstance(msg_id, basestring):
			raise ValueError, "msg_id must be a string"
		if not isinstance(read, bool):
			raise ValueError, "read must be a boolean True/False"

		#Get database
		conn, cur = self.get_database()

		#Check if the mark read action is in database
		cur.execute("SELECT read FROM mark_read_actions WHERE msg_id = ? LIMIT 1", [msg_id])
		readstate = cur.fetchone()
		if readstate == None:	#Insert if not there
			cur.execute("INSERT INTO mark_read_actions (msg_id, read) VALUES (?,?)", [msg_id, read])
		elif readstate[0] != read:	#Update if doesn't match
			cur.execute("UPDATE mark_read_actions SET read = ? WHERE msg_id = ?", [read, msg_id])

		#Mark it in database
		self._set_read_state(msg_id, read, cur)

		#Return database
		self.return_database(conn, cur)

	def search_messages(self, keywords):
		#TODO: Lazily load extension or use LIKE
		pass

	def send_message(self, message):
		"""Send a message
			message: Message to transmit, must be an instance of email.message.Message

			Remarks: message ought to have the from address set, it must have the to address set
		"""
		#Check parameters
		if not isinstance(message, email.message.Message):
			raise ValueError, "Message is not an instance of email.message.Message"
		if message["To"] == None:
			raise ValueError, "Message does not have a to address!"
		if message["From"] == None:
			message["From"] = self.username

		#As of now this works, we'll just let it die miserably if it fails :)
		server = smtplib.SMTP("smtp.gmail.com", 587)
		code, data = server.ehlo()
		server.starttls()
		server.ehlo()
		server.login(self.username, self.password)
		server.sendmail(message["From"], message["To"], message.as_string())
		server.quit()

		"""	A typical smtp session:
		>>> import smtplib
		>>> server = smtplib.SMTP("smtp.gmail.com", 587)
		>>> server.ehlo()
		(250, 'mx.google.com at your service, [212.10.154.179]\nSIZE 35651584\n8BITMIME\nSTARTTLS\nENHANCEDSTATUSCODES')
		>>> server.starttls()
		(220, '2.0.0 Ready to start TLS')
		>>> server.ehlo()
		(250, 'mx.google.com at your service, [212.10.154.179]\nSIZE 35651584\n8BITMIME\nAUTH LOGIN PLAIN\nENHANCEDSTATUSCODES')
		>>> server.login("testraptormail@gmail.com","testtest")
		(235, '2.7.0 Accepted')
		>>> server.sendmail(m["From"], m["To"], m.as_string())
		{}
		>>> server.quit()
		"""
		#TODO: Intepret and handle reply codes, they're documented here: http://tools.ietf.org/html/rfc5321#page-51


"""
Notes:
[Gmail], [Gmail]/Spam [Gmail]/Trash are not labels as these are dynamic



Tasks:
 - Add label
 - Remove label
 - Delete message
 - Star message
 - Send message
 - View message
 - Search content
 - List labels
 - List label, order, limit, data-to-return
 - List conversation, limit
 - Create draft message	(Or just SaveDraftMessage)
 - Edit draft message	(Or just SaveDraftMessage)


See: http://tools.ietf.org/html/rfc2822

sync:
 - Perform archived actions


 - List labels and add nonexisting
 - Loop through labels (except. dynamic ones [Gmail], [Gmail]/Spam and [Gmail]/Trash)
	 - UID fetch <last_uid>:* (BODY[HEADER.FIELDS (message-id)])
		 - Check if msg_id exists in mails, fetch message and add if not
		 - Add msg_id relation to label_relations, if not there
		 - Update last_uid for the label




 - Fetch new message messages
	- select "[Gmail]/All mail"
	- UID fetch <last_uid>:* (rfc822)		<last_uid> = 1 or lastuid from settings
	- parse mails and add to database, if they're not in it, identify them by Message-ID
	- Note the last UID and save it to settings
 - Sync labels, for each label <label>
	- select <label>
	- UID fetch <last_uid>:* (BODY[HEADER.FIELDS (message-id)])
				<last_uid> = 1 or lastuid from labeltable
	- 
	
"""

import threading

class AsyncSyncer(threading.Thread):
	def __init__ ( self, gmail):
		self._gmail = gmail
		threading.Thread.__init__ (self)

	def run (self):
		print "Syncronization initiated"
		self._gmail.sync()
		print "Syncronization complete"

def G(tosync = True):
	gmail = GMail("testraptormail@gmail.com","testtest", "testdb.rmdb")
	if tosync:
		AsyncSyncer(gmail).start()
	return gmail


