#!/usr/bin/env python
# -*- coding: utf-8 -*-

import string
import struct
import numbertheory
class IDEA:
	"""	Class implementing IDEA symmetric encryption
		Algorithms by Handbook of Applied Cryptography and Applied Cryptography

	>>> i = IDEA(1231242342234)
	>>> c = i.encrypt("my text of unspecified length")
	>>> i.decrypt(c)
	'my text of unspecified length'
	"""
	def __init__(self, key, mode = "ECB", impl = "auto"):
		"""	Creates an instance of IDEA
			key: encryption key, as integer or string
			mode: ECB or CBC, operation mode
			impl: managed, native or auto
		"""
		#Check for valid operation mode
		if mode not in ['ECB','CBC']:
			raise ValueError, "Mode of operation not supported"
		self._mode = mode
		if impl not in ["managed", "native", "auto"]:
			raise ValueError, "implementation must be managed, native or auto"
		#Detect algorithm implementation
		if impl == "auto":
			try:
				import idea_optimized
				self._optimized = idea_optimized
				self._impl = "native"
			except ImportError:		#If we get an import error let's assume native implementation is unavailable
				self._impl = "managed"
		elif impl == "native":
			import idea_optimized	#Let it raise an error if user demands native implementation
			self._optimized = idea_optimized
			self._impl = "native"
		else:
			self._impl = "managed"

		if not isinstance(key, int) and not isinstance(key, long):
			raise ValueError, "Key must be of an integer or string"
		self._key = self._compute_subkeys(key)
		self._inverse_key = None	#Lazy key generation
		pass

	def _process_block(self, block, keyset):
		""" Process a block
			By Handbook of Applied Cryptography, Algorithm 7.101
			block: Datablock to encrypt, will be subject to modification
			keyset: keyset to encrypt with
			Returns the encrypted block
		"""
		data = list(block)
		for r in range(0,8):
			#(a)
			data[1-1] = self._mod_mul(data[1-1], keyset[1-1+6*r])
			data[4-1] = self._mod_mul(data[4-1], keyset[4-1+6*r])
			data[2-1] = (data[2-1] + keyset[2-1+6*r]) & 0xffff
			data[3-1] = (data[3-1] + keyset[3-1+6*r]) & 0xffff
			#(b)
			t0 = self._mod_mul(keyset[5-1+6*r], data[1-1] ^ data[3-1])
			t1 = self._mod_mul(keyset[6-1+6*r], (t0 + (data[2-1] ^ data[4-1])) & 0xffff)
			t2 = (t0 + t1) & 0xffff
			#(c)
			data[1-1] = data[1-1] ^ t1
			data[4-1] = data[4-1] ^ t2
			a = data[2-1] ^ t2
			data[2-1] = data[3-1] ^ t1
			data[3-1] = a
		Y = []
		Y += [self._mod_mul(data[1-1], keyset[1-1+6*8])]
		Y += [(data[3-1] + keyset[2-1+6*8]) & 0xffff]
		Y += [(data[2-1] + keyset[3-1+6*8]) & 0xffff]
		Y += [self._mod_mul(data[4-1], keyset[4-1+6*8])]
		return Y

	def encrypt(self, data):
		"""	Encrypts data using initialized key, mode and implementation
			Returns decrypted (binary) string
		"""
		data = self._pad_data(data)
		if self._impl == "native":
			return self._optimized.encrypt(data, self._mode, self._key)
		else:
			return self._encrypt(data)

	def decrypt(self, data):
		"""	Decrypts data using initialized key, mode and implementation
			Returns decrypted (binary) string
		"""
		#Generate inverse keys lazily
		if self._inverse_key == None:
			self._inverse_key = self._compute_inverse_subkeys(self._key)
		if self._impl == "native":
			return self._unpad_data(self._optimized.decrypt(data, self._mode, self._inverse_key))
		else:
			return self._unpad_data(self._decrypt(data))

	def _encrypt(self, data):
		""" Encrypt data using managed implementation
			Returns encrypted binary string
		"""
		blocks = self._break_data(data)
		cblocks = []
		if self._mode == 'ECB':
			for block in blocks:
				cblocks += [self._process_block(block, self._key)]
		else:
			cblocks += [[3,14,15,92]]
			for block in blocks:
				xblock = [cblocks[-1][0]^block[0],cblocks[-1][1]^block[1],cblocks[-1][2]^block[2],cblocks[-1][3]^block[3]]
				cblocks += [self._process_block(xblock,self._key)]
		return self._join_blocks(cblocks)

	def _decrypt(self, data):
		"""	Decrypts data using managed implementation
			Returns decrypted (binary) string
		"""
		cblocks = self._break_data(data)
		blocks = []
		if self._mode == 'ECB':
			for cblock in cblocks:
				blocks += [self._process_block(cblock, self._inverse_key)]
		else:
			pcblock = cblocks[0]
			for cblock in cblocks[1:]:
				block = self._process_block(cblock, self._inverse_key)
				blocks += [[pcblock[0]^block[0],pcblock[1]^block[1],pcblock[2]^block[2],pcblock[3]^block[3]]]
				pcblock = cblock
		return self._join_blocks(blocks)

	def _pad_data(self, data):
		"""	Pad data length to be a multiple of 64bit
		"""
		missing = (8 -((len(data)+1)%8))%8
		data += struct.pack("B", missing+1)
		data += '\0'*(missing)
		return data

	def _unpad_data(self, data):
		"""	Remove padding added by the _pad_data method
			Raises exception if data wasn't padded properly
		"""
		padding = -1
		while data[padding] == '\0':
			padding -= 1
		[missing] = struct.unpack("B", data[padding])
		if -missing != padding:
			raise ValueError, "Data was not probably padded"
		return data[:padding]

	def _break_data(self, data):
		"""	Breake data into blocks
		"""
		if len(data)%8:
			raise ValueError, "Data must be a multiple of 64bits"
		blocks = []
		i = 0
		while i < len(data):
			blocks += [struct.unpack("H"*4, data[i:i+8])]
			i+=8
		return blocks

	def _join_blocks(self, blocks):
		"""	Joins blocks to a string
		"""
		data = ""
		for block in blocks:
			data += struct.pack("H"*4, block[0], block[1], block[2], block[3])
		return data

	def _compute_subkeys(self, key):
		"""	Computes the subkeys of key
		"""
		keys = []
		for x in [-1]*6+[3]:
			for i in range(7,x,-1):
				keys += [(key & 0xffff<<16*i) >> 16*i]
			key = (key<<25 & (2**128-1)) | key>>(128-25)
		return keys

	def _compute_inverse_subkeys(self, keys):
		"""	Computes inverse subkey of a key
		"""
		if not isinstance(keys, list):
			raise ValueError, "key must be a list of subkeys"
		ikeys = []
		ikeys += [self._invert_mod_n(keys[1-1+6*(9-1)])]
		ikeys += [(2**16 - keys[2-1+6*(9-1)]) & 0xffff]
		ikeys += [(2**16 - keys[3-1+6*(9-1)]) & 0xffff]
		ikeys += [self._invert_mod_n(keys[4-1+6*(9-1)])]
		ikeys += [keys[5-1+6*(8-1)]]	
		ikeys += [keys[6-1+6*(8-1)]]
		for r in range(2,9):
			ikeys += [self._invert_mod_n(keys[1-1+6*(9-r)])]
			ikeys += [(2**16 - keys[3-1+6*(9-r)]) & 0xffff]
			ikeys += [(2**16 - keys[2-1+6*(9-r)]) & 0xffff]
			ikeys += [self._invert_mod_n(keys[4-1+6*(9-r)])]
			ikeys += [keys[5-1+6*(8-r)]]
			ikeys += [keys[6-1+6*(8-r)]]
		ikeys += [self._invert_mod_n(keys[1-1+6*(9-9)])]
		ikeys += [(2**16 - keys[2-1+6*(9-9)]) & 0xffff]
		ikeys += [(2**16 - keys[3-1+6*(9-9)]) & 0xffff]
		ikeys += [self._invert_mod_n(keys[4-1+6*(9-9)])]
		return ikeys

	def _invert_mod_n(self, k):
		"""	Calculates the multiplicative inverse of k mod (2**16+1)
		"""
		d, x, y = numbertheory.ext_gcd(2**16+1, k)
		if y < 0:
			return 2**16+1 + y
		return y

	def _mod_mul(self, a, b):
		""" Modified multiplication of a and b
		"""
		if a == 0:
			a = 2**16
		if b == 0:
			b = 2**16
		r = (a*b) % ((2**16)+1)
		if r == 2**16:
			return 0
		return r
