#!/usr/bin/env python
# -*- coding: utf-8 -*-

import math
#import gmpy
import os
import struct
import sys

def mod_exp(a, k, n):
	"""
		Modular exponentiation
		Returns: a**k mod n
		Implementation of repeated square and multiply for exponentiation in Zn
		 from Handbook of Applied Cryptography, algorithm 2.147 page 71

		>>> mod_exp(433, 73, 77) == 20
		True
		>>> mod_exp(33, 555, 98) == 13
		True
		>>> mod_exp(33, 553, 98) == 18
		False
	"""
	if k < 0:
		a = invert(a,n)	
		k = -k
	# b is set to 1 which is the result if k is 0 ie. a**0 = 1
	b = 1
	if k == 0:
		return b
	# A is set to a to preserve the value a. 
	A = a
	for i in range(0, bitlength(abs(k))):
		if k & (1 << i):
			b = (A * b) % n
		A = (A * A) % n
	return b	 


def bitlength(k):
	"""
		Returns the bitlength of k

		The largest number we can represent with
		  a bitlength x is 2**x-1.

		Use logarithm to isolate x:
		2**x-1 = k
		x*log(2) - log(1) = log(k)
		x = log(k+1) / log(2)

		>>> bitlength(500) == 9
		True
		>>> bitlength(50) == 6
		True
		>>> bitlength(5) == 4
		False
	"""
	
	return int(math.ceil(math.log(k+1,2)))


def gcd(a, b):
	"""
		Returns the greatest common divisor for a and b

		>>> gcd(888,777) == 111
		True
		>>> gcd(8232,723) == 3
		True
		>>> gcd(823,777) == 5
		False
	"""
	if a < b:
		a,b =b,a
	d = a
	while b!= 0:
		t=b
		b= d % b
		d = t
	return d
	#return gmpy.gcd(a,b)


def ext_gcd(a, b):
	"""
		Compute: d == a * x + b * y
		Return: d, x, y

		>>> ext_gcd(8, 9) == (1, -1, 1)
		True
		>>> ext_gcd(434, 3453) == (1, -1090, 137)
		True
		>>> a, b = 43, 56
		>>> d, x, y = ext_gcd(a, b)
		>>> d == a * x + b * y
		True
	"""
	if b == 0:
		return a, 1, 0
	x1, x2 = 0, 1
	y1, y2 = 1, 0
	while b > 0:
		q = a / b
		r, x, y = a - q * b, x2 - q * x1, y2 - q * y1
		a = b
		b = r
		x2 = x1
		x1 = x
		y2 = y1
		y1 = y
	d, x, y = a, x2, y2
	return d, x, y

#Sources:
#Handbook of Applied Cryptography
#http://www.cs.utoronto.ca/~trebla/ExtendedEuclid.txt
#Latter is a real nice explanation

def invert(a, n):
	"""
		Compute multiplicated invers of a mod n in Zn

	"""
	
	d, x, y = ext_gcd(a, n)
	if d != 1: 
		raise ValueError, "a and n are not coprime, multiplicative inverse of a mod n does not exist"
	return x


def random_number(a):
	"""Returns a random number less than a"""
	k = bitlength(a)
	if k%8 == 0:
		k = k/8
	else:
		k = k/8 +1
	r = struct.unpack("B"*k, os.urandom(k))
	R = 1
	for i in range(0, k):
		R = ((2**(i*8))*r[i] + R)%a
	return R

def is_prime(n, t = 30):
	"""	Test if a is a prime """
	if n == 2: return True
	if n%2==0 or n == 1: return False
	r = n-1
	s = 0
	while r%2 == 0:
		r = r/2
		s +=1
	for i in range(0,t):
		a = random_number(n-2)+2
		y = mod_exp(a,r,n)
		if y != 1 and y != n-1:
			j = 1
			while j < s and y != n-1:
				y = y**2 % n
				if y == 1: return False
				j += 1
			if y != n-1: return False
	return True


def previous_prime(a, t = 30):
	"""Get a prime smaller than a with the certainty of t"""
	if a <= 2:
		raise ValueError, "Can't find any primes less than 2"
	b = a
	if b % 2 == 0: b-=1
	while not is_prime(b,t): b-=2
	return b

def next_prime(a, t = 30):
	"""Get a prime larger than a with the certainty of t"""
	if a < 2: return 2
	b = a
	while not is_prime(b,t): b+=2
	return b

def random_prime(n, t = 30):
	"""Get a random prime smaller than n with a certainty of t"""
	return previous_prime(random_number(n-3)+3,t)

def testp():
	for i in range(2,20):
		ds = gmpy.is_prime(i)
		vs = is_prime(i,25)
		if ds == 2 or ds == 1:
			if not vs:
				print "Missed prime:\t" + str(i)
		else:
			if vs:
				print "Got not prime:\t" + str(i)

def _test():
	import doctest
	doctest.testmod()

if __name__ == "__main__":
	_test()


